#!/usr/bin/env python

"""
lab.py
Copyright (C) 2007  Yannick Le Teigner

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import gtk
import gobject
import math

class Connection(object):
    """Class representing a connection between two devices' port"""

    def __init__(self, x, y, devicea=None):
        """Creates a new connection"""
        self.pxa = self.pxb = self.xa = self.xb = x
        self.pya = self.pyb = self.ya = self.yb = y
        self.devicea = devicea
        self.deviceb = None
        self.__selected = False

    def __str__(self):
        """Return a string representation of this connection"""
        return "%s --/ %s" % (self.devicea, self.deviceb)

    def delete(self):
        """Delete this connection from the lab"""
        self.devicea.delete_connection(self)
        self.deviceb.delete_connection(self)

    def _get_selected(self):
        """Set this connection to user selected mode or not"""
        return self.__selected

    def _set_selected(self, value):
        """Set this connection to user selected mode or not"""
        self.__selected = value

    selected = property(_get_selected, _set_selected, doc='is this connection currently selected by the user')

    def move0to(self, x, y):
        """Move a side of the connection to (x, y)"""
        self.pxa = self.xa
        self.pya = self.ya
        self.xa = x
        self.ya = y

    def move1to(self, x, y):
        """Move b side of the connection to (x, y)"""
        self.pxb = self.xb
        self.pyb = self.yb
        self.xb = x
        self.yb = y

    def translate0to(self, tx, ty):
        """Translate a side of the connection by [x, y]"""
        self.pxa = self.xa
        self.pya = self.ya
        self.xa += tx
        self.ya += ty

    def translate1to(self, tx, ty):
        """Translate b side of the connection by [x, y]"""
        self.pxb = self.xb
        self.pyb = self.yb
        self.xb += tx
        self.yb += ty

    def draw(self, context):
        """Draw this connection on the canvas"""
        if self.selected:
            context.set_source_rgb(255, 0, 0)
        else:
            context.set_source_rgb(0, 0, 0)
        context.move_to(self.xa, self.ya)
        context.line_to(self.xb, self.yb)
        context.stroke()


    def intersect(self, area):
        """Returns true if this connection intersects area"""
        left, top, right, bottom = area
        l = min(self.xa, self.xb)
        r = max(self.xa, self.xb)
        t = min(self.ya, self.yb)
        b = max(self.ya, self.yb)
        if left < l: left = l
        if top < t: top = t
        if right > r: right = r
        if bottom > b: bottom = b
        return not (left >= right or top >= bottom)

    def dirty_area(self):
        """Returns a (x0, y0, x1, y1) dirty area """
        minx = min(60000, self.xa, self.pxa, self.xb, self.pxb)
        maxx = max(0, self.xa, self.pxa, self.xb, self.pxb)
        miny = min(60000, self.ya, self.pya, self.yb, self.pyb)
        maxy = max(0, self.ya, self.pya, self.yb, self.pyb)
        return (minx-2, miny-2, maxx+2, maxy+2) # increase area due to AA

    def inside(self, x, y):
        """Returns true if (x,y) is on the connection"""
        if ((x > min(self.xa, self.xb)-2) and (x < max(self.xa, self.xb)+2)
                and (y > min(self.ya, self.yb)-2)
                and (y < max(self.ya, self.yb)+2)):
            a = self.xb - self.xa
            b = self.yb - self.ya
            d = abs(a*(self.ya-y) - (self.xa-x)*b) / math.sqrt(a*a + b*b)
            return d < 4.0
        return False


class Lab(object):
    """Class representing a Lab object.

    A Lab object contains a list of the devices and connections of a given
    topology.

    """

    def __init__(self):
        """Create a new lab"""
        self.devices = []      # list of devices on the canvas
        self.device = None     # device that is currently selected
        self.connections = []  # list of connections
        self.connection = None # the connection that is being drawn
        self.selected_ccts = []

    def reset(self):
        """Reset the state of this lab"""
        self.devices = []      # list of devices on the canvas
        self.device = None     # device that is currently selected
        self.connections = []  # list of connections
        self.connection = None # the connection that is being drawn

# singleton
lab = Lab()
