#!/usr/bin/env python

"""
lab_io.py
Copyright (C) 2006  Yannick Le Teigner

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import re
import os
import sys
import gtk
from gui_c7200 import GuiC7200, GuiPA_FE_TX, GuiPA_4Tp, GuiPA_8T, GuiPA_A1, GuiC7200_IO_FE, GuiC7200_IO_2FE, GuiPA_4E, GuiPA_8E
from gui_c3600 import GuiC3600, GuiLeopard_2FE, GuiNM_1FE_TX, GuiNM_1E, GuiNM_4E, GuiNM_4T, GuiNM_16ESW
from gui_router import GuiRouter, dynamips
from gui_frsw import GuiFRSwitch
from gui_atmsw import GuiATMSwitch
from gui_ethsw import GuiEthSwitch
from lab import lab, Connection
from dynamips_lib import Dynamips, DynamipsError, NIO_gen_eth, NOSEND
from util import RouterSelectionDialog
import dynamips_lib

interface_re = re.compile(r"^[f|a|s|e|p](\d\/\d+)$", re.IGNORECASE)
nio_re = re.compile(r"^(NIO_gen_eth):(.*)$")
mapping_re = re.compile(r"^\d+:\d+$")
vpc_re = re.compile(r"^(\d+):(\d+)[:/]?([\d]*)$")
number_re = re.compile(r"^\d+$")
slot_re = re.compile(r"^slot(\d)$", re.IGNORECASE)

def load_lab(filename, use_running=False):
    # load the gui part
    from configobj import ConfigObj, flatten_errors
    if os.path.exists(filename):
        try:
            config = ConfigObj(filename, raise_errors=True)
        except SyntaxError, e:
            print "Error:", e
            sys.exit(1)
    else:
        print "Lab file doesn't exist: ", filename
        return
    cct = []        # connections between router, frsw
    mappings = []   # mappings inside a frsw
    lab.reset()
    config_tab = {} # router to saved config hash
    default_values = {} # default values specified in the lab file

    for key in config.keys():
        # this is global parameter
        if type(config[key])==str:
            print 'not supported yet: ', key
            continue

        # Sections should be dynamips instances
        # Create a dynagen instance and add it to the list
        if ':' in key: # the format is [ip:port]
            host, port = key.split(':')
            port = int(port)
        else:
            host = key
            port = int(config[key].get('port', 7200))
        try:
            d = Dynamips(host, port)
        except DynamipsError:
            print "Cannot connect to the lab: ", host, port
            # FIXME add parent
            gtk.MessageDialog(None, gtk.DIALOG_DESTROY_WITH_PARENT,
                              gtk.MESSAGE_ERROR, gtk.BUTTONS_OK,
                              "Cannot connect to the dynamips server (%s:%d)."\
                                      % (host, port))
            # FIXME cleanup
            continue # go to the next dynamips instance, if any
        # some global parameters for the dynamips instance
        wdir = config[key].get('workingdir', None)
        if wdir is not None:
            d.workingdir = wdir
        udp = config[key].get('udp', None)
        if udp is not None:
            d.udp = int(udp)
            d.base_udp = d.udp
        print dynamips_lib.NOSEND
        if use_running:
            dynamips_lib.NOSEND = True
        else:
            d.reset()
        dynamips.append(d)
        default_values[d] = {}

        # go through all the devices for this dynagen instance
        for section in config[key].sections:
            if section in ['7200', '3640']: #FIXME load defaults here
                default_values[d][section] = config[key][section]
            else:
                device = config[key][section]
                devtype, name = device.name.split(' ')
                if devtype.lower()=='router':
                    try:
                        x = int(device['x'])
                        y = int(device['y'])
                    except:
                        import random
                        x = random.randint(0, 300)
                        y = random.randint(0, 250)
                    try:
                        model = device['model']
                    except:
                        model = None
                    if model=='7200' or model is None:
                        print dynamips_lib.NOSEND
                        r = GuiC7200(dynamips=d, x=x, y=y, name=name)
                    elif model in ['3620', '3640', '3660']:
                        r = GuiC3600(dynamips=d, x=x, y=y, name=name,
                                     chassis=model)
                    # fill in default values
                    if model in default_values[d].keys():
                        def_val = default_values[d][model]
                        for att in def_val.keys():
                            val = def_val[att]
                            #if att == 'image':
                            #    val = '"'+val+'"'
                            if att in ('ram', 'disk0', 'disk1'):
                                val = int(val)
                            elif att in ('mmap', 'sparsemem'):
                                val = val in ('true', 'True')
                            r.__setattr__(att, val)
                    # fill in specific values
                    for k in device.keys():
                        if k=='image':
                            image = device[k]
                            if os.path.exists(image): # FIXME remote images
                                r.image = image
                            else:
                                print "Image %s does not exist, or not local" % image
                                r.image = image
                        elif k=='ram':
                            r.ram = int(device[k])
                        elif k=='sparsemem':
                            r.sparsemem = device[k] in ('True', 'true')
                        elif k=='console':
                            r.console = int(device[k])
                        elif k=='configuration':
                            r.set_tmp_config_gz64(device[k])
                            config_tab[r] = device[k]
                            #r.set_config_gz64(device[k])
                        elif k=='idlepc':
                            r.idlepc = device[k]
                        elif k=='npe':
                            r.npe = device[k]
                        elif k=='midplane':
                            r.midplane = device[k]
                        elif k=='cnfg':
                            r.cnfg = '"'+device[k]+'"'
                        elif k=='colorbg_r':
                            r.colorbg_r = float(device[k])
                        elif k=='colorbg_g':
                            r.colorbg_g = float(device[k])
                        elif k=='colorbg_b':
                            r.colorbg_b = float(device[k])
                        elif slot_re.match(k):
                            slot_number = int(slot_re.findall(k)[0])
                            pa_type = device[k]
                            if pa_type=='C7200-IO-FE' and slot_number==0:
                                r.slot[0]=GuiC7200_IO_FE(r, 0)
                            elif pa_type=='C7200-IO-2FE' and slot_number==0:
                                r.slot[0]=GuiC7200_IO_2FE(r, 0)
                            elif pa_type=='PA-FE-TX':
                                r.slot[slot_number] = GuiPA_FE_TX(r, slot_number)
                            elif pa_type=='PA-4T+' or pa_type=='PA-4T':
                                r.slot[slot_number] = GuiPA_4Tp(r, slot_number)
                            elif pa_type=='PA-8T':
                                r.slot[slot_number] = GuiPA_8T(r, slot_number)
                            elif pa_type=='PA-A1':
                                r.slot[slot_number] = GuiPA_A1(r, slot_number)
                            elif pa_type=='PA-4E':
                                r.slot[slot_number] = GuiPA_4E(r, slot_number)
                            elif pa_type=='PA-8E':
                                r.slot[slot_number] = GuiPA_8E(r, slot_number)
                            # 36x0 PAs
                            elif pa_type=='NM-1E':
                                r.slot[slot_number] = GuiNM_1E(r, slot_number)
                            elif pa_type=='NM-1FE-TX':
                                r.slot[slot_number] = GuiNM_1FE_TX(r, slot_number)
                            elif pa_type=='NM-4T':
                                r.slot[slot_number] = GuiNM_4T(r, slot_number)
                            elif pa_type=='NM-4E':
                                r.slot[slot_number] = GuiNM_4E(r, slot_number)
                            elif pa_type=='NM-16ESW':
                                r.slot[slot_number] = GuiNM_16ESW(r, slot_number)
                        elif interface_re.match(k):
                            interface = interface_re.findall(k)[0]
                            if nio_re.match(device[k]): # remote is NIO
                                niotype, niostring = nio_re.findall(device[k])[0]
                                remote_device = niotype
                                remote_port = niostring
                            else:
                                (remote_device, remote_port) = device[k].split(' ')
                            cct.append([r, interface, remote_device, remote_port])
                    # for compatibility with files that have no "slot0:" section
                    if isinstance(r, GuiC7200) and r.slot[0]==None:
                        r.slot[0] = GuiC7200_IO_FE(r, 0)
                    lab.devices.append(r)
                elif devtype.lower()=='frsw':
                    try:
                        x = int(device['x'])
                        y = int(device['y'])
                    except:
                        x = y = 10
                    frsw = GuiFRSwitch(dynamips=d, x=x, y=y, name=name)
                    for k in device.keys():
                        if k=='colorbg_r':
                            frsw.colorbg_r = float(device[k])
                        elif k=='colorbg_g':
                            frsw.colorbg_g = float(device[k])
                        elif k=='colorbg_b':
                            frsw.colorbg_b = float(device[k])
                        elif number_re.match(k):
                            (remote_device, remote_port) = device[k].split(' ')
                            cct.append([frsw, int(k), remote_device, remote_port])
                        elif slot_re.match(k):
                            slot_number = int(slot_re.findall(k)[0])
                        elif mapping_re.match(k):
                            pa, dlcia = map(int, k.split(':'))
                            pb, dlcib = map(int, device[k].split(':'))
                            mappings.append([frsw, pa, dlcia, pb, dlcib])
                    lab.devices.append(frsw)
                elif devtype.lower()=='atmsw':
                    try:
                        x = int(device['x'])
                        y = int(device['y'])
                    except:
                        x = y = 10
                    atmsw = GuiATMSwitch(dynamips=d, x=x, y=y, name=name)
                    for k in device.keys():
                        if k=='colorbg_r':
                            atmsw.colorbg_r = float(device[k])
                        elif k=='colorbg_g':
                            atmsw.colorbg_g = float(device[k])
                        elif k=='colorbg_b':
                            atmsw.colorbg_b = float(device[k])
                        elif number_re.match(k):
                            (remote_device, remote_port) = device[k].split(' ')
                            cct.append([atmsw, int(k), remote_device, remote_port])
                        elif slot_re.match(k):
                            slot_number = int(slot_re.findall(k)[0])
                        elif vpc_re.match(k):
                            conn = vpc_re.findall(k)[0]
                            pa = int(conn[0])
                            if len(conn)>2:
                                vpca = conn[1]+':'+conn[2]
                            else:
                                vpca = conn[1]
                            conn = vpc_re.findall(device[k])[0]
                            pb = int(conn[0])
                            if len(conn)>2:
                                vpcb = conn[1]+':'+conn[2]
                            else:
                                vpcb = conn[1]
                            mappings.append(
                                    [atmsw, int(pa), vpca, int(pb), vpcb])
                    lab.devices.append(atmsw)
                elif devtype.lower()=='ethsw':
                    try:
                        x = int(device['x'])
                        y = int(device['y'])
                    except:
                        x = y = 10
                    ethsw = GuiEthSwitch(dynamips=d, x=x, y=y, name=name)
                    for k in device.keys():
                        if k=='colorbg_r':
                            ethsw.colorbg_r = float(device[k])
                        elif k=='colorbg_g':
                            ethsw.colorbg_g = float(device[k])
                        elif k=='colorbg_b':
                            ethsw.colorbg_b = float(device[k])
                        elif number_re.match(k):
                            (a, b) = device[k].split(' ')
                            # FIXME
                            if a.lower() in ['access', 'dot1q']:
                                # add port definition
                                ethsw.addport(int(k), int(b),
                                              a.lower()=='dot1q')
                            else:
                                # add connection
                                cct.append([ethsw, int(k), a, b])
                    lab.devices.append(ethsw)
    for cc in cct:
        devicea = cc[0]
        deviceb = None

        # FIXME cleanup
        #if cc[2].lower() == 'nio_gen_eth':
        #    slot, pa = map(int, cc[1].split('/'))
        #    devicea.slot[slot].nio(pa, nio=NIO_gen_eth(devicea.dynamips, interface=cc[3]))
        #    continue

        for r in lab.devices:
            if r.name == cc[2]:
                deviceb = r
                break
        if deviceb is None:
            print "Error, no device with the name '%s'" % cc[2]
            continue
        if isinstance(devicea, GuiRouter):
            slot, pa = map(int, cc[1].split('/'))
            (xa, ya) = devicea.getPortPos(slot, pa)
            connection = Connection(xa, ya, devicea)
            if isinstance(deviceb, GuiRouter):
                cc[3] = interface_re.findall(cc[3])[0]
                rslot, rpa = map(int, cc[3].split('/'))
                (connection.xb, connection.yb) = deviceb.getPortPos(rslot, rpa)
                connection.deviceb = deviceb
            elif isinstance(deviceb, (GuiFRSwitch, GuiATMSwitch, GuiEthSwitch)):
                rport = int(cc[3])
                (connection.xb, connection.yb) = deviceb.getPortPos(rport)
                connection.deviceb = deviceb
            else:
                print 'error in connection'
                continue
            devicea.addconnection(connection, 0)
            deviceb.addconnection(connection, 1)
        elif isinstance(devicea, GuiFRSwitch):
            rport = cc[1]
            (xa, ya) = devicea.getPortPos(rport)
            connection = Connection(xa, ya, devicea)
            if isinstance(deviceb, GuiRouter):
                cc[3] = interface_re.findall(cc[3])[0]
                rslot, rpa = map(int, cc[3].split('/'))
                (connection.xb, connection.yb) = deviceb.getPortPos(rslot, rpa)
                connection.deviceb = deviceb
                deviceb.addconnection(connection, 1)
            devicea.addconnection(connection, 0)
        elif isinstance(devicea, GuiATMSwitch):
            rport = cc[1]
            (xa, ya) = devicea.getPortPos(rport)
            connection = Connection(xa, ya, devicea)
            if isinstance(deviceb, GuiRouter):
                cc[3] = interface_re.findall(cc[3])[0]
                rslot, rpa = map(int, cc[3].split('/'))
                (connection.xb, connection.yb) = deviceb.getPortPos(rslot, rpa)
                connection.deviceb = deviceb
                deviceb.addconnection(connection, 1)
            devicea.addconnection(connection, 0)
        lab.connections.append(connection)
    for m in mappings:
        m[0].addmapping(m[1], m[2], m[3], m[4])

    if len(config_tab.keys()):
        text = 'Below devices have their configuration\nsaved in the lab file.\nDo you want to push them to be activated?'
        rcd = RouterSelectionDialog(config_tab.keys(), text)
        # need to push config for below devices
        for d in rcd.selected_devices:
            print 'trying to push'
            d.config_gz64 = config_tab[d]
            print 'pushed'
            d.push_cfg = True

    if use_running:
        dynamips_lib.NOSEND = False

def save_lab(filename):
    output = "# Lab - created by dynagui.\n"

    # try to extract as many configs as possible
    saved_config = {}
    for device in [x for x in lab.devices if isinstance(x, GuiRouter)]:
        if device.started is False \
                and device.stopped_config64bz is not None: #saved before stopped
            config = device.stopped_config64bz
            saved_config[device] = config
        else: # or retrieve it now
            try:
                config = device.config_gz64
                saved_config[device] = config
            except:
                pass

    # remove the config the user does not want to save
    keys = saved_config.keys()
    if not (keys == []):
        text = 'Please select the devices for which you would like\nto save their configuration into the lab file.'
        rcd = RouterSelectionDialog(keys, text)
        selected = rcd.selected_devices
        for d in keys:
            if not (d in selected):
                del saved_config[d]

    # save each device
    for d in dynamips:
        if d.port != 7200:
            output += "[%s:%d]\n" % (d.host, d.port)
        else:
            output += "[%s]\n" % (d.host)
        output += "workingdir = %s\n" % d.workingdir
        if 'base_udp' in dir(d):
            output += "udp = %d\n" % d.base_udp
        #output += "\n    [[7200]]\n"  # default for all routers
        for device in lab.devices:
            if device.dynamips is d:
                output = __save_device(output, device, saved_config)
        output += "\n"

    file = open(filename, 'w')
    file.write(output)
    file.close()


def __save_device(output, device, saved_config):
    # mapping of 'key name' to [default value, type]
    keys = {'x': [0, int],
            'y': [0, int],
            'colorbg_r': [0, float],
            'colorbg_g': [0, float],
            'colorbg_b': [0, float],
            'ram': [256, int],
            'disk0': [64, int],
            'disk1': [64, int],
            'cnfg' : [None, str]}
    if isinstance(device, GuiRouter): # common attributes among routers
        output += "\n    [[ROUTER %s]]\n" % device.name
        if isinstance(device, GuiC3600):
            output += "    model = %s\n" % device.chassis
        if device in saved_config.keys():
            output += "    configuration = %s\n" % saved_config[device]
        if device.image is not None:
            output += "    image = \"%s\"\n" % device.image
        if device.sparsemem:
            output += "    sparsemem = True\n"
        if device.idlepc is not None:
            output += "    idlepc = \"%s\"\n" % device.idlepc
        output += "    console = %d\n" % device.console
        for k in keys.keys():
            if k in dir(device):
                val = getattr(device, k)
                if val!=keys[k][0]:
                    output += "    %s = %s\n" % (k, getattr(device, k))

    if isinstance(device, GuiC7200):
        if device.npe != 'npe-200':
            output += "    npe = %s\n" % device.npe
        if device.midplane != 'vxr':
            output += "    midplane = %s\n" % device.midplane
        for j in range(0, len(device.slot)):
            if device.slot[j]!=None:
                output += "    slot%d = %s\n" % (j, device.slot[j].adapter)
        for cctl in device.connections:
            pa, port, cct, side = cctl
            if side==0: # save the connection if we are '0' side
                deviceb = cct.deviceb
                if isinstance(deviceb, GuiRouter):
                    r_pa_slot, xl, yl = deviceb.which_pa(cct.xb, cct.yb)
                    r_pa = deviceb.slot[r_pa_slot]
                    r_port = r_pa.which_port(xl, yl)
                    output += "    %s%d/%d = %s %s%d/%d\n"\
                        % (device.slot[pa].technology, pa, port, deviceb.name,\
                        r_pa.technology, r_pa_slot, r_port)
                elif isinstance(deviceb, GuiFRSwitch):
                    r_port = deviceb.which_port(cct.xb, cct.yb)
                    output += "    %s%d/%d = %s %d\n"\
                        % (device.slot[pa].technology, pa, port, deviceb.name,\
                        r_port)
                elif isinstance(deviceb, GuiATMSwitch):
                    r_port = deviceb.which_port(cct.xb, cct.yb)
                    output += "    %s%d/%d = %s %d\n"\
                        % (device.slot[pa].technology, pa, port, deviceb.name,\
                        r_port)
                elif isinstance(deviceb, GuiEthSwitch):
                    r_port = deviceb.which_port(cct.xb, cct.yb)
                    output += "    %s%d/%d = %s %d\n"\
                        % (device.slot[pa].technology, pa, port, deviceb.name,\
                        r_port)
            elif side==1 and isinstance(cct.devicea, GuiEthSwitch):
                r_port = cct.devicea.which_port(cct.xa, cct.ya)
                output += "    %s%d/%d = %s %d\n"\
                    % (device.slot[pa].technology, pa, port, cct.devicea.name,\
                    r_port)

    elif isinstance(device, GuiC3600):
        if device.chassis in ['3620', '3640']:
            index = 0
        elif device.chassis == '3660':
            index = 1
        for j in range(index, len(device.slot)):
            if device.slot[j]!=None:
                output += "    slot%d = %s\n" % (j, device.slot[j].adapter)
        for cctl in device.connections:
            pa, port, cct, side = cctl
            if side==0: # save the connection if we are '0' side
                deviceb = cct.deviceb
                if isinstance(deviceb, GuiRouter):
                    r_pa_slot, xl, yl = deviceb.which_pa(cct.xb, cct.yb)
                    r_pa = deviceb.slot[r_pa_slot]
                    r_port = r_pa.which_port(xl, yl)
                    output += "    %s%d/%d = %s %s%d/%d\n"\
                        % (device.slot[pa].technology, pa, port, deviceb.name,\
                        r_pa.technology, r_pa_slot,r_port)
                elif isinstance(deviceb, GuiFRSwitch):
                    r_port = deviceb.which_port(cct.xb, cct.yb)
                    output += "    %s%d/%d = %s %d\n"\
                        % (device.slot[pa].technology, pa, port, deviceb.name,\
                        r_port)
                elif isinstance(deviceb, GuiEthSwitch):
                    r_port = deviceb.which_port(cct.xb, cct.yb)
                    output += "    %s%d/%d = %s %d\n"\
                        % (device.slot[pa].technology, pa, port, deviceb.name,\
                        r_port)
            elif side==1 and isinstance(cct.deviceb, GuiEthSwitch):
                r_port = deviceb.which_port(cct.xa, cct.ya)
                output += "    %s%d/%d = %s %d\n"\
                        % (device.slot[pa].technology, pa, port, devicea.name,\
                        r_port)

    elif isinstance(device, GuiFRSwitch):
        output += "\n    [[FRSW %s]]\n" % device.name
        output += "    x = %d\n" % device.x
        output += "    y = %d\n" % device.y
        output += "    colorbg_r = %f\n" % device.colorbg_r
        output += "    colorbg_g = %f\n" % device.colorbg_g
        output += "    colorbg_b = %f\n" % device.colorbg_b
        for m in device.mappings:
            output += "    %d:%d = %d:%d\n" % (m[0], m[1], m[2], m[3])
        for cctl in device.connections:
            port, cct, side = cctl
            if side==0: # save the connection if we are '0' side
                deviceb = cct.deviceb
                if isinstance(deviceb, GuiRouter):
                    r_pa_slot, xl, yl = deviceb.which_pa(cct.xb, cct.yb)
                    r_pa = deviceb.slot[r_pa_slot]
                    r_port = r_pa.which_port(xl, yl)
                    output += "    %d = %s %s%d/%d\n"\
                        % (port, deviceb.name, r_pa.technology, r_pa_slot,\
                        r_port)
                elif isinstance(deviceb, GuiFRSwitch):
                    r_port = deviceb.which_port(cct.xb, cct.yb)
                    output += "    %d = %s %d\n" % (port, deviceb.name, r_port)

    elif isinstance(device, GuiATMSwitch):
        output += "\n    [[ATMSW %s]]\n" % device.name
        output += "    x = %d\n" % device.x
        output += "    y = %d\n" % device.y
        output += "    colorbg_r = %f\n" % device.colorbg_r
        output += "    colorbg_g = %f\n" % device.colorbg_g
        output += "    colorbg_b = %f\n" % device.colorbg_b
        for m in device.mappings:
            output += "    %d:%s = %d:%s\n" % (m[0], m[1], m[2], m[3])

        for cctl in device.connections:
            port, cct, side = cctl
            if side==0: # save the connection if we are '0' side
                deviceb = cct.deviceb
                if isinstance(deviceb, GuiRouter):
                    r_pa_slot, xl, yl = deviceb.which_pa(cct.xb, cct.yb)
                    r_pa = deviceb.slot[r_pa_slot]
                    r_port = r_pa.which_port(xl, yl)
                    output += "    %d = %s %s%d/%d\n"\
                        % (port, deviceb.name, r_pa.technology, r_pa_slot,\
                        r_port)
                elif isinstance(deviceb, GuiATMSwitch):
                    r_port = deviceb.which_port(cct.xb, cct.yb)
                    output += "    %d = %s %d\n" % (port, deviceb.name, r_port)


    elif isinstance(device, GuiEthSwitch):
        output += "\n    [[ethsw %s]]\n" % device.name
        output += "    x = %d\n" % device.x
        output += "    y = %d\n" % device.y
        output += "    colorbg_r = %f\n" % device.colorbg_r
        output += "    colorbg_g = %f\n" % device.colorbg_g
        output += "    colorbg_b = %f\n" % device.colorbg_b

        for i, m in enumerate(device.ports):
            if not (m == [None, None]):
                if m[1] is True:
                    output += "    %d = dot1q %s\n" % (i, m[0])
                else:
                    output += "    %d = access %s\n" % (i, m[0])
        for i, m in enumerate(device.unusedports):
            if not (m == [None, None]):
                if m[1] is True:
                    output += "    %d = dot1q %s\n" % (i, m[0])
                else:
                    output += "    %d = access %s\n" % (i, m[0])
    #4 = dot1q 1 NIO_gen_eth:eth0
    #4 = dot1q 1 NIO_gen_eth:\Device\NPF_{B00A38DD-F10B-43B4-99F4-B4A078484487}

    return output

